# -*- coding: utf-8 -*-
"""
Created on Wed Mar 19 09:04:02 2025

@author: Shagedoorn1
"""
import numpy as np
import sympy as sp
import matplotlib.pyplot as plt
from transfer_function import TransferFunction
def bode_plot(transfer_function, omega_range=np.logspace(-1, 6, 1000)):
    """
    Generate and display the Bode plot of the given system.
    
    A Bode plot consists of 2 subplots:
        1 magnitude (in decibels) vs frequency (in radians per second)
        2 phase (in degrees) vs frequency (in radians per second)
    
    Parameters:
        transfer_function (TransferFunction):
            A TransferFunction object, see transfer_function.py for details
        omega_range (numpy array):
            A range of frequencies for which the Bode plot is computed,
            default from 10^-1 to 10^6, with 1000 points.
    """
    magnitude, phase = transfer_function.calc_magnitude_and_phase(omega_range)
    
    plt.figure(figsize = (10, 6))
    plt.subplot(2, 1, 1)
    plt.semilogx(omega_range, magnitude)
    plt.title("Bode plot")
    plt.ylabel("Magnitude (dB)")
    plt.axhline(linewidth=1, color='Black')
    plt.grid(True)
    
    plt.subplot(2, 1, 2)
    plt.semilogx(omega_range, phase)
    plt.ylabel("phase (deg)")
    plt.xlabel("Frequency (rad/s)")
    plt.axhline(linewidth=1, color='Black')
    plt.axhline(y=-90, linewidth=1, color="Black")
    plt.grid(True)
    
    plt.show()

def pz_map(transfer_function):
    """
    Generate and display the Pole-Zero map of the given system.
    
    The map shows the poles and zeros of the transfer function in the complex
    plane
        - Poles are marked with an "X"
        - Zeros are marked with an "O"
    
    Parameters:
        transfer_function (TransferFunction):
            The transfer_function object that represents the system,
            see transfer_function.py for details.
    
    """
    poles = transfer_function.poles
    zeros = transfer_function.zeros
    i = 0
    j = 0
    reals = []
    comps = []
    plt.title("Pole-Zero map")
    plt.scatter(
        [sp.re(p) for p in poles],
        [sp.im(p) for p in poles],
        marker = 'x',
        s = 500,
        color = 'black',
        label = 'poles')
    while i < len(zeros):
        reals.append(sp.re(zeros[i]))
        comps.append(sp.im(zeros[i]))
        i += 1
    plt.scatter(
        [sp.re(z) for z in zeros],
        [sp.im(z) for z in zeros],
        marker = "o",
        s = 500,
        edgecolors = "black",
        facecolors = 'none',
        label = 'zeros')
    while j < len(poles):
        reals.append(sp.re(poles[j]))
        comps.append(sp.im(poles[j]))
        j += 1
    plt.xlim([float(min(reals)) - 1, float(max(reals)) + 1])
    plt.ylim([float(min(comps))-1, float(max(comps)) + 1])
    plt.axhline(linewidth=1, color='Black')
    plt.axvline(linewidth=1, color='Black')
    plt.xlabel(r"$\lambda$ re(s)")
    plt.ylabel(r"$j \omega$ im(s)")
    plt.legend(prop={'size': 15})
    plt.grid()
    plt.show()
    
if __name__ == "__main__":
    string1 = "y''+y'+y"
    string2 = "x'+x"
    o_range = np.logspace(-1,3,1000)
    Trans = TransferFunction(string1, string2)
    bode_plot(Trans, omega_range=np.logspace(-1,3,1000))
    pz_map(Trans)


from .Diffeq import DiffEq
from .laplace_handler import LaplaceHandler
from .transfer_function import TransferFunction
from .plotting import bode_plot, pz_map

__all__ = ["DiffEq", "LaplaceHandler", "TransferFunction", "bode_plot", "pz_map"]


# -*- coding: utf-8 -*-
"""
Created on Tue Mar 18 10:53:55 2025

@author: Shagedoorn1
"""
import sympy as sp
import numpy as np
import math
from sympy import I
from laplace_handler import LaplaceHandler
from Diffeq import DiffEq


class TransferFunction:
    """
    A class that calculates the transfer function of a control system from its
    governing ordinary differential equations, ODEs.
    
    Two ODEs should be provided one for the input signal and one
    for the output signal. These are laplace transformed and set equal to eachother.
    
    The transfer function H(s) is defined as the ratio:
        H(s) = Y(s) / X(s)
    
    If a more readable version of the equations is desired, omit print 
    statements and directly evaluate the required value in the editor or
    console.
    """
    
    def __init__(self, equation_out, equation_in, name_out='y', name_in='x', var='t'):
        """
        Initializes the TransferFunction class.
        
        Parameters:
            equation_out (str):
                The ODE representing the output signal
            equation_in (str):
                The ODE representing the input signal
            name_out (str):
                The name of the output function, default y
            name_in (str):
                The name of the input function, defaulte x
            var (str):
                The independent variable of the ODEs, default t
        
        Attributes:
            laplace_transform_out (sympy expression):
                The laplace transform of the output equation
            laplace_transform_in (sympy expression):
                The laplace transform of the input equation
            transfer_func (sympy expression):
                The calculated transfer function H(s)
            num (sympy expression):
                This is the numerator of the transfer function
            den (sympy expression):
                The denominator of the transfer functions
            zeros (list):
                The zeros of the transfer function, these are the roots of the
                numerator
            poles (list):
                The poles of the transfer function, these are the roots of the
                denominator
            
        """
        self.equation_out = equation_out
        self.equation_in = equation_in
        
        self.name_out = name_out
        self.name_in = name_in
        
        self.var = var
        
        self.lap_out = LaplaceHandler(self.equation_out, self.name_out, self.var)
        self.lap_in = LaplaceHandler(self.equation_in, self.name_in, self.var)
        
        self.y = self.lap_out.name
        self.x = self.lap_in.name
        self.s = sp.Symbol('s')
        self.w = sp.Symbol('w')
        
        self.Y = sp.Function(self.name_out.upper())(self.s)
        self.X = sp.Function(self.name_in.upper())(self.s)
        self.H = sp.Function('H')
        
        self.domain = True
        
        self.out_func = DiffEq(self.equation_out, self.name_out, self.var).function
        self.in_func = DiffEq(self.equation_in, self.name_in, self.var).function
        
        self.laplace_transform_out = self.lap_out.laplace_transform
        self.laplace_transform_in = self.lap_in.laplace_transform
        
        self.transfer_func = self.compute_h()
        
        self.num, self.den = sp.fraction(self.transfer_func)
        self.zeros, self.poles = self.compute_zeros_and_poles()
    def compute_h(self, print_eq=False):
        """
        Computes the transfer function H(s).
        
        This function subtracts the Laplace transformed equations from each
        other, sets the result equal to zero and solves for Y(s).
        Finally, H(s) is computed by dividing by X(s)
        
        Parameters:
            print_eq (bool):
                If True, prints the equation before solving, default False.
        Returns:
            Hs (sympy expression):
                The transfer function H(s).
        """
        eq = self.laplace_transform_out - self.laplace_transform_in
        if print_eq:
            sp.pprint(f"{eq} = 0")
        H = sp.solve(eq, self.Y)[0] #remove list
        Hs = H/self.X
        return Hs
    
    def compute_zeros_and_poles(self):
        """
        Computes the zeros and poles of the transfer function.
        
        Zeros are found by setting the numerator equal to 0, poles are found
        by setting the denominator equal to 0.
        
        Returns:
            list:
                zeros
                poles
        """
        zeros = sp.solve(self.num)
        poles = sp.solve(self.den)
        return zeros, poles
    
    def switch_domain(self):
        """
        Switches the domain of the transfer function between Laplace (s-domain)
        and Fourier (w-domain).
        
        Replace s with jw, Laplace-to-Fourier,
        or w with -js, Fourier-to-Laplace
        Returns:
            Sympy expression:
                The transfer function in the switched domain.
        """
        if self.domain:
            self.transfer_func = self.transfer_func.subs(self.s, (I*self.w)) #Laplace to Fourier
        elif not self.domain:
            self.transfer_func = self.transfer_func.subs((self.w), -I*self.s) #Fourier to Laplace
        self.domain = not self.domain
        return self.transfer_func
    
    def transfer_function_numeric(self):
        """
        Converts the symbolic transfer function into a numeric function.
        
        Ensures the function is in the Fourier domain before numeric substitution
        
        Returns:
            Function:
                A function that evaluates the transfer function at specified
                values of omega.
        """
        if self.domain: #Make sure to be in Fourier domain
            self.switch_domain()
        
        def H_numeric(omega):
            transfer_func = self.transfer_func.subs(self.w, omega) #Substitute w for numbers
            return transfer_func
        return H_numeric
    
    def calc_magnitude_and_phase(self, omega_range = np.logspace(-1,6,1000)):
        """
        Calculates the magnitude and phase of the transfer function over a
        given range of frequencies.
        
        The magnitude is expressed in decibels using:
            magnitude = 20 * log10(|H(jw)|)
        
        The phase is expressed in degrees using:
            phase = arctan(re(H)/im(H) * (180/pi))
        
        Parameter:
            omega_range:
                A range of frequencies, in rad/s, over which the magnitude and
                phase are computed.
                Default is logarithmically spaced range from 10^-1 to 10^6 with
                1000 points.
        
        Returns:
            lists:
                magnitude: The magnitude response in decibels
                phase: The phase response in degrees
        """
        H_numeric = self.transfer_function_numeric()
        
        magnitude = []
        phase = []
        
        for omega in omega_range:
            H = H_numeric(omega)
            magnitude.append(20 * np.log10(float(abs(H))))
            phase.append((180/math.pi) * math.atan((sp.im(H))/(sp.re(H))))
        return magnitude, phase
    
    def __str__(self):
        """
        Returns a formatted string representation of the system.
        
        Displays the systems differential equation, its Laplace transform and
        the transfer function.
        """
        return f"ODE: {self.out_func} = {self.in_func} \n Laplace transform: {self.laplace_transform_out} = {self.laplace_transform_in} \n H(s) = {self.transfer_func}"

if __name__ == '__main__':
    string1 = "y''+1*y'+y"
    string2 = "x+x'"
        
    Trans = TransferFunction(string1, string2)
    print(Trans.transfer_func)


# -*- coding: utf-8 -*-
"""
Created on Fri Mar 14 18:58:04 2025

@author: Shagedoorn1
"""
import sympy as sp
from Diffeq import DiffEq
from numpy import zeros
sp.init_printing()

class LaplaceHandler():
    """
    A class for correctly computing the laplace transform of a differentail equation.
    
    This class properly handles the transformation of an ordinary differential
    equation (ODE) into the Laplace domain, ensuring the correct handling of
    functions and initial conditions.
    """
    def __init__(self, equation, name, var='t'):
        """
        Initializes the LaplaceHandler class.
        
        Parameters:
            equation (str):
                The differential equation to be transformed
            name:
                The name of the function in the equation, commonly 'y'
            var:
                The independent variable of the equation, default 't'
        
        Attributes:
            t (sympy Symbol):
                The independent variable before transforming
            s (sympy Symbol):
                The complex-valued laplace variable
            name (sympy Function):
                The function to be transformed
            lap_name (sympy Function):
                The name of the laplace-transformed function,
                capitalized version of 'name'.
            function (DiffEq object):
                A symbolic expression of the input equation (see Diffeq.py)
            equation:
                The differential equation converted to a sympy expression.
            initial_conds (numpy array):
                Initial conditions of the differential equations.
                In control theory, these are assumed to be zero, meaning the
                system is at rest before the input signal is applied.
            laplace_transform (sympy expression):
                The laplace transformed version of the input equation
        """
        self.t = sp.Symbol(var.lower())
        self.s = sp.Symbol('s')
        
        self.name = sp.Function(name.lower())
        self.lap_name = sp.Function(name.upper())
        self.function = DiffEq(equation, name, var)
        self.equation = self.function.function
        
        if self.function.order == 0:
            self.initial_conds = [0]
        else:
            self.initial_conds = zeros(self.function.order)
        self.laplace_transform = self.transform()
    
    def transform(self):
        """
        This transforms the given equation into the Laplace domain.
        
        This function applies the Laplace transform to the input equation using
        sympy's built in functions, while ensuring correct handling of initial
        conditions and function substitution.
        
        Returns:
            sympy expression:
                The Laplace-transformed equation.
        """
        transformed = sp.laplace_transform(self.equation, self.t, self.s, noconds=True)
        transformed = sp.laplace_correspondence(transformed, {self.name: self.lap_name})
        transformed = sp.laplace_initial_conds(transformed, self.t, {self.name: self.initial_conds})
        return transformed
    
    def __str__(self):
        """
        Returns the Laplace-transformed equation as a formatted string.
        
        When an instance of this class is printed, the transformed equation is
        displayed in a readable format
        """
        sp.pretty_print(f"{self.laplace_transform}")
        return ""
if __name__ == '__main__':
    string = "y'''+y''+y'+y+sin(t)"
    a = LaplaceHandler(string, 'y')
    print(a)